#!/bin/sh
# Overlay text demo
# Use "-d /dev/videoX" to select board

./overlay $@ -q -i 1 -x 80 -y 80 -t ff7f7f "Location 45.420657,-122.752284"
./overlay $@ -q -i 2 -x 520 -y 80 -t 7fff7f "Temperature 65F"
./overlay $@ -q -i 3 -x 80 -y 400 -t ffff7f "Windspeed 5mph"
./overlay $@    -i 4 -x 520 -y 400 -t 7f7fff "Humidity 10%"
