#!/bin/sh
# Overlay lines demo
# Use "-d /dev/videoX" to select board

# clear the whole screen
./overlay $@ -q -i 0 -l  0 0  0 575 720 ff000000

# draw a diamond in the center of the display
./overlay $@ -i 1 -l  \
	100 240  360 120  1 ffffff	\
	360 120  620 240  1 ffffff	\
	620 240  360 360  1 ffffff	\
	360 360  100 240  1 ffffff

# draw a 50% alpha filled square in the center of the display
./overlay $@ -i 2 -l 340 220  340 260  40 7f7fffff
