#!/bin/sh
# Overlay font demo
# Use "-d /dev/videoX" to select board

# clear the whole screen
./overlay $@ -q -l  0 0  0 575 720 ff000000

#FONT=/usr/share/fonts/truetype/freefont/FreeSerif.ttf
#FONT=/usr/share/fonts/truetype/freefont/FreeSans.ttf
FONT=FreeSans

./overlay $@ -i 1 -x 30 -y 50 -f ${FONT} -s 64 -o -S -t ffffff "Sensoray Co., Inc."

./overlay $@ -i 2 -x 30 -y 150 -f ${FONT} -s 64 -o -S -I -t 00ff00 -a 1 "Sensoray Co., Inc."

./overlay $@ -i 3 -x 30 -y 250 -f ${FONT} -s 64 -o -S -b -t ff0000 -a 2 "Sensoray Co., Inc."

./overlay $@ -i 4 -x 30 -y 350 -f ${FONT} -s 64 -o -S -b -I -t 4040ff -a 3 "Sensoray Co., Inc."

