#!/bin/sh
# Overlay images demo
# Use "-d /dev/videoX" to select board

./overlay $@ -q -i 1 -x 50 -y 30   Sensoray_Emb_Elec.png
./overlay $@ -q -i 2 -x 480 -y 20  Compass.png
./overlay $@ -q -i 3 -x 50 -y 200  Gauges.png
./overlay $@    -i 4 -x 380 -y 250 Weather.png

